GRAPH = (("file_name",),{'type':str,'help':"use graph file GRAPH",'metavar':'GRAPH'})
CACHE = (("file_name",),{'type':str,'help':"use cache file CACHE",'metavar':'CACHE'})
SOURCE = (("-s","--source"),{'type':int,'dest':'source_id','help':"set source power tower to SOURCE (default: random)",'metavar':'SOURCE'})
TARGET = (("-t","--target"),{'type':int,'dest':'target_id','help':"set target power tower to TARGET (default: random)",'metavar':'TARGET'})
OUT = (("-o","--output"),{'type':str,'dest':'out_file','help':"write path to OUT (default: stdout)",'metavar':'OUT'})
INDIRECT = (("-i","--indirect"),{'action':'store_true','dest':'indirect','default':False,'help':"transform indirectly (default: False)"})
PENALIZE = (("-p","--penalize"),{'type':int,'default':20,'dest':'penalize','help':"penalize air time by factor PENALTY",'metavar':'PENALTY'})
VISUALIZE = (("-m","--map"),{'action':'store_true','dest':'visualize','default':False,'help':"render route to HTML map"})

CONFIG = [
    ("npz",{'help':"route using NPZ graph",'args':[INDIRECT,PENALIZE,SOURCE,TARGET,OUT,VISUALIZE,GRAPH]}),
    ("nx",{'help':"route using NX graph",'args':[SOURCE,TARGET,OUT,VISUALIZE,GRAPH]}),
    ("gt",{'help':"route using GT graph",'args':[INDIRECT,PENALIZE,SOURCE,TARGET,OUT,VISUALIZE,GRAPH]}),
    ("direct",{'help':"route directly",'args':[
        (("-s","--source"),{'type':int,'dest':'source_id','help':"set source power tower to SOURCE (default: random)",'metavar':'SOURCE','required':True}),
        (("-t","--target"),{'type':int,'dest':'target_id','help':"set target power tower to TARGET (default: random)",'metavar':'TARGET','required':True}),
        OUT,
        (("-u","--overpass-url"),{'type':str,'dest':'overpass_url','help':"define url for Overpass API to be URL (default: http://caracal.imada.sdu.dk/)",'metavar':'URL'}),
        (("-d","--disk-cache"),{'action':'store_true','dest':'disk_cache','default':False,'help':"use disk cache (default: False)"}),
        VISUALIZE,
        CACHE
    ]})
]

def astar_path_npz(lat, long, id2edges, edges_weight, edges_neighbor, source, target):
    from itertools import count
    from heapq import heappush, heappop
    from limic.util import haversine_distance
    targetlong = long[target]
    targetlat = lat[target]
    c = count()
    queue = [(0, next(c), source, 0, None)]
    enqueued = {}
    explored = {}
    while queue:
        _, __, curnode, dist, parent = heappop(queue)
        if curnode == target:
            path = [curnode]
            node = parent
            while node is not None:
                path.append(node)
                node = explored[node]
            path.reverse()
            return path
        if curnode in explored:
            if explored[curnode] is None:
                continue
            qcost, h = enqueued[curnode]
            if qcost < dist:
                continue
        explored[curnode] = parent
        left = id2edges[curnode]
        right = id2edges[curnode+1]
        for i in range(left,right):
            w = edges_weight[i]
            neighbor = edges_neighbor[i]
            ncost = dist + w
            if neighbor in enqueued:
                qcost, h = enqueued[neighbor]
                if qcost <= ncost:
                    continue
            else:
                h = haversine_distance(long[neighbor], lat[neighbor], targetlong, targetlat)
            enqueued[neighbor] = ncost, h
            heappush(queue, (ncost + h, next(c), neighbor, ncost, curnode))
    raise ValueError("node %s not reachable from %s" % (target, source))

def astar_npz(g,source_id=None,target_id=None):
    from limic.util import locate_by_id, haversine_distance
    ids = g['ids']
    lat = g['lat']
    long = g['long']
    id2edges = g['id2edges']
    edges_weight = g['edges_weight']
    edges_neighbor = g['edges_neighbor']
    edges_air = g['edges_type']
    source, target = locate_by_id(ids,source_id,target_id)
    try:
        path = astar_path_npz(lat, long,id2edges,edges_weight,edges_neighbor,source,target)
    except ValueError as e:
        path = []
    def index2node(cost,dist,air,index):
        return (cost,dist,air,ids[index],lat[index],long[index])
    if path:
        cost = dist = 0.0
        dpath = [index2node(cost,dist,False,path[0])]
        for i in range(1,len(path)):
            left = id2edges[path[i]]
            right = id2edges[path[i]+1]
            for j in range(left,right):
                if edges_neighbor[j] == path[i-1]:
                    dist = edges_weight[j]
                    air = edges_air[j]
                    cost += dist
                    dpath.append(index2node(cost,dist,air < 0,path[i]))
                    break
        return cost,dpath
    else:
        dist = haversine_distance(long[source],lat[source],long[target],lat[target])
        return None,[index2node(0.,0.,False,source),index2node(float('inf'),dist,True,target)]

def route_npz(file_name,source_id=None,target_id=None,out_file=None,indirect=False,penalize=20,visualize=False):
    from limic.util import start, end, status, save_path, load_npz
    from limic.convert import transform_npz_nx
    from numpy import int32
    start("Loading from",file_name)
    g = load_npz(file_name)
    end()
    start("Checking whether GT graph is rescaled")
    if g['edges_weight'].dtype == int32:
        indirect = True
        status("YES (forcing routing through NX)")
    else:
        status("NO")
    if indirect:
        start("Transforming graph to NX format")
        h = transform_npz_nx(g,penalize)
        end()
        start("Routing using NX")
        path = astar_nx(h,source_id,target_id)
        end()
    else:
        start("Routing using NPZ")
        path = astar_npz(g,source_id,target_id)
        end()
    start("Saving path to",out_file)
    save_path(path,out_file,visualize)
    end()

def astar_nx(g,source_id=None,target_id=None):
    from limic.util import locate_by_id, haversine_distance
    from networkx import astar_path, NetworkXNoPath
    if isinstance(source_id,tuple) and isinstance(target_id,tuple):
        source, target = source_id, target_id
    else:
        nodes = list(g.nodes())
        ids = list(map(lambda x:x[0],nodes))
        source, target = locate_by_id(ids,source_id,target_id)
        source, target = nodes[source], nodes[target]
    targetlat, targetlong = target[1], target[2]
    def distance(x,y):
        #assert(target==y)
        return haversine_distance(longx=x[2],latx=x[1],longy=targetlong,laty=targetlat)
    try:
        path = astar_path(g,source,target,heuristic=distance)
        cost = dist = 0.0
        dpath = [(cost,dist,False)+path[0]]
        for i in range(1,len(path)):
            ggg = g[path[i-1]][path[i]]
            dist = ggg['weight']
            cost += dist
            air = ggg['type']
            dpath.append((cost,dist,air < 0)+path[i])
        return cost,dpath
    except NetworkXNoPath as e:
        cost = dist = distance(source,target)
        path = None,[(0.,0.,False)+source,(float('inf'),dist,True)+target]
    return path

def route_nx(file_name,source_id=None,target_id=None,out_file=None,visualize=False):
    from limic.util import start, end, load_pickled, save_path
    start("Loading from",file_name)
    g = load_pickled(file_name)
    end()
    start("Routing using NX")
    path = astar_nx(g,source_id,target_id)
    end()
    start("Saving path to",out_file)
    save_path(path,out_file,visualize)
    end()

def astar_gt(g,source_id=None,target_id=None):
    from graph_tool.search import astar_search, AStarVisitor, StopSearch
    from limic.util import locate_by_id, haversine_distance
    long = g.vp.long
    lat = g.vp.lat
    ids = list(map(lambda x:x[1],g.get_vertices([g.vp.id])))
    source, target = locate_by_id(ids,source_id,target_id)
    source, target = g.vertex(source), g.vertex(target)
    targetlong, targetlat = long[target], lat[target]
    class Visitor(AStarVisitor):
        def __init__(self, target):
            self.target = target
        def edge_relaxed(self, e):
            if e.target() == self.target:
                raise StopSearch()
    def distance(x):
        return haversine_distance(longx=long[x],latx=lat[x],longy=targetlong,laty=targetlat)
    dist, pred = astar_search(g, source, g.ep.weight, Visitor(target), heuristic=distance)
    path = []
    current = target
    total_cost = cost = dist[current]
    if g.vertex(pred[current]) == current:
        air = True
        dist = distance(source)
    else:
        e = g.edge(g.vertex(pred[current]),g.vertex(current))
        dist = g.ep.weight[e]
        air = True if g.ep.type[e] < 0 else False
    while True:
        i = int(current)
        path.append((cost,dist,air < 0,ids[i],lat[i],long[i]))
        pre = g.vertex(pred[current])
        if pre == current:
            break
        cost -= dist
        current = pre
        pre = g.vertex(pred[current])
        if pre == current:
            dist = 0.
            air = False
        else:
            e = g.edge(g.vertex(pre),g.vertex(current))
            dist = g.ep.weight[e]
            air = True if g.ep.type[e] < 0 else False
    path.reverse()
    if len(path) < 2:
        assert(len(path)==1)
        source = int(source)
        return None if source != target else 0.,[(0.,0.,False,ids[source],lat[source],long[source])]+(path if source != target else [])
    else:
        return total_cost,path

def route_gt(file_name,source_id=None,target_id=None,out_file=None,indirect=False,penalize=20,visualize=False):
    from limic.util import start, end, status, load_gt, save_path
    from limic.convert import transform_gt_npz
    start("Loading graph from",file_name)
    g = load_gt(file_name)
    end()
    start("Checking whether GT graph is rescaled")
    if g.gp.rescaled:
        indirect = True
        status("YES (forcing routing through NPZ)")
    else:
        status("NO")
    if indirect:
        start("Transforming graph to NPZ format")
        h = transform_gt_npz(g,penalize)
        end()
        start("Routing using NPZ")
        path = astar_npz(h,source_id,target_id)
        end()
    else:
        start("Routing using GT")
        path = astar_gt(g,source_id,target_id)
        end()
    start("Saving path to",out_file)
    save_path(path,out_file,visualize)
    end()

def reconstruct_path_direct(current):
    from limic.util import distance
    dist = current.g-current.origin.g if current.origin else 0.
    air = distance(current.tower.latlon,current.origin.tower.latlon) < 0.9*dist if current.origin else False
    cost = 0.
    path = [(current.g,dist,air,current.tower.id,current.tower.latlon[0],current.tower.latlon[1])]
    while current.origin:
        current = current.origin
        cost = current.g
        dist = cost-current.origin.g if current.origin else 0.
        air = distance(current.tower.latlon,current.origin.tower.latlon) < 0.9*(current.g-current.origin.g) if current.origin else False
        path.append((cost,dist,air,current.tower.id,current.tower.latlon[0],current.tower.latlon[1]))
    path.reverse()
    return cost, path

def astar_tower_direct(start,end,max_fly=1000,eps=0.01,safe_dist=100,penalty=20,prec=1):
    class node:
        def __init__(self,f,g,tower,origin):
            self.f = f
            self.g = g
            self.tower = tower
            self.origin = origin
        def __lt__(self,other):
            return self.f < other.f
        def __repr__(self):
            return "node(f="+repr(self.f)+",g="+repr(self.g)+",tower="+repr(self.tower)+",origin="+repr(self.origin.tower.id if self.origin else self.origin)+")"
    from limic.overpass import find_all_neighbours
    from heapq import heappop, heappush, heapify
    from limic.util import distance
    from math import inf
    latlon2id = {}
    minusid = [0]
    start_node = node(distance(start.latlon,end.latlon)*prec,0,start,None)
    nodes = {start.id : start_node}
    todo = [start_node]
    while todo:
        #if verbosity >= 4: print("*******************\n"+str(todo)+"\n***************")
        #if verbosity >= 4: print("*******************\n"+str(sorted(todo))+"\n***************")
        current = heappop(todo)
        #if verbosity >= 2: print("visiting",current.f,current.tower.id,current.tower.latlon)
        #if verbosity >= 2: print("full node",current)
        if current.tower.id == end.id:
            return reconstruct_path_direct(current)
        #if verbosity >=2:
        #    print("CURRENT PATH:\nnode(id:"+(",".join(map(str,reconstruct_path(current)[1])))+");out;")
        #    print("FRONT:\nnode(id:"+(",".join(map(lambda n:str(n.tower.id),todo)))+");out;")
        neighbours = find_all_neighbours(current.tower, max_fly, eps, safe_dist, minusid, latlon2id, penalty)
        for neighbour in neighbours:
            g = current.g+neighbour.dist
            neighbour_node = nodes.get(neighbour.tower.id,node(None,inf,neighbour.tower,None))
            if g < neighbour_node.g:
                if not neighbour_node.f:
                    nodes[neighbour.tower.id] = neighbour_node
                must_add = neighbour_node not in todo
                neighbour_node.origin = current
                neighbour_node.g = g
                neighbour_node.f = g+distance(neighbour.tower.latlon,end.latlon)*prec
                if must_add:
                    #if verbosity >= 3: print("added",neighbour_node)
                    heappush(todo,neighbour_node)
                #else:
                    #if verbosity >= 3: print("modified",neighbour_node)
                heapify(todo)
    return None,[(0.,0.,False,start.id,start.latlon[0],start.latlon[1]),(float('inf'),start_node.f,True,end.id,end.latlon[0],end.latlon[1])]

def astar_direct(start,end):
    from limic.overpass import tower_by_id
    start_tower = tower_by_id(start)
    end_tower = tower_by_id(end)
    return astar_tower_direct(start_tower,end_tower)
    
def route_direct(file_name,source_id=None,target_id=None,out_file=None,overpass_url=None,disk_cache=False,visualize=False):
    from limic.util import start, end, status, file_size, load_pickled, save_pickled, save_path, options, replace
    if disk_cache:
        start("Using disk cache",file_name)
        set_option('disk_cache',file_name)
    from limic.overpass import region, set_server
    if disk_cache:
        status("OK")
    from os.path import exists
    if not disk_cache and exists(file_name):
        start("Loading",file_name)
        region.backend._cache = load_pickled(file_name)
        end('')
        file_size(file_name)
    len_cache = len(region.backend._cache)
    start("Routing using direct algorithm")
    set_server(overpass_url)
    path = astar_direct(source_id,target_id)
    end()
    start("Saving path to",out_file)
    save_path(path,out_file,visualize)
    end()
    if not disk_cache and len_cache != len(region.backend._cache):
        file_name_tmp = file_name+".tmp"
        start("Saving to",file_name,"via",file_name_tmp)
        save_pickled(file_name_tmp,region.backend._cache)
        replace(file_name_tmp,file_name)
        end('')
        file_size(file_name)



IN = (("file_name_in",),{'type':str,'help':"graph file to convert",'metavar':'GRAPH'})
OUT = (("file_name_out",),{'type':str,'help':"graph file to save to",'metavar':'GRAPH'})
INC = (("file_name_in",),{'type':str,'help':"cache file to convert",'metavar':'CACHE'})
OUTC = (("file_name_out",),{'type':str,'help':"cache file to save to",'metavar':'CACHE'})
INDIRECT = (("-i","--indirect"),{'action':'store_true','dest':'indirect','default':False,'help':"transform indirectly via GT"})
RESCALE = (("-r","--rescale"),{'action':'store_true','default':False,'dest':'rescale','help':"rescale to int32"})
PENALIZE = (("-p","--penalize"),{'type':int,'default':20,'dest':'penalize','help':"penalize air time by factor PENALTY",'metavar':'PENALTY'})
CONFIG = [
    ("nx",{'help':"convert NX graph",'args':[
        ("gt",{'help':"convert to GT graph",'args':[
            RESCALE,IN,OUT
        ]}),
        ("npz",{'help':"convert to NPZ graph",'args':[
            INDIRECT,RESCALE,PENALIZE,IN,OUT
        ]}),
    ]}),
    ("gt",{'help':"convert GT graph",'args':[
        ("nx",{'help':"convert to NX graph",'args':[
            PENALIZE,IN,OUT
        ]}),
        ("npz",{'help':"convert to NPZ graph",'args':[
            PENALIZE,IN,OUT
        ]})
    ]}),
    ("npz",{'help':"convert NPZ graph",'args':[
        ("nx",{'help':"convert to NX graph",'args':[
            PENALIZE,IN,OUT
        ]}),
        ("gt",{'help':"convert to GT graph",'args':[
            RESCALE,PENALIZE,IN,OUT
        ]})
    ]}),
    ("cache",{'help':"convert pickled cache",'args':[
        ("dbm",{'help':"convert to DBM cache",'args':[
            INC,OUTC
        ]})
    ]}),
    ("dbm",{'help':"convert DBM cache",'args':[
        ("cache",{'help':"convert to pickled cache",'args':[
            INC,OUTC
        ]})
    ]})
]

def transform_nx_gt(g,rescale=False):
    from graph_tool import Graph
    from limic.util import haversine_distance
    h = Graph(directed=False)
    h.vp.id = h.new_vertex_property("int64_t")
    if rescale:
        h.vp.lat = h.new_vertex_property("int32_t")
        h.vp.long = h.new_vertex_property("int32_t")
        h.ep.weight = h.new_edge_property("int32_t")
    else:
        h.vp.lat = h.new_vertex_property("double")
        h.vp.long = h.new_vertex_property("double")
        h.ep.weight = h.new_edge_property("float")
    h.ep.type = h.new_edge_property("int32_t")
    h.gp.rescaled = h.new_graph_property("bool")
    h.gp.rescaled = rescale
    pos2vertex = {}
    intersection_id = 0
    for n in g.nodes():
        v = h.add_vertex()
        pos2vertex[n[1],n[2]] = v
        if n[0] < 0:
            intersection_id -= 1
            h.vp.id[v] = intersection_id
        else:
            h.vp.id[v] = n[0]
        h.vp.lat[v] = int(n[1]*10000000) if rescale else n[1]
        h.vp.long[v] = int(n[2]*10000000) if rescale else n[2]
    for n,m,d in g.edges.data(data=True):
        w = d['weight']
        air = d['type']
        e = h.add_edge(pos2vertex[n[1],n[2]],pos2vertex[m[1],m[2]])
        h.ep.type[e] = air
        if rescale and air < 0:
            w = haversine_distance(longx=n[2],latx=n[1],longy=m[2],laty=m[1])
        h.ep.weight[e] = int(w*1000) if rescale else w
    return h

def convert_nx_gt(file_name_in,file_name_out,rescale=False):
    from limic.util import start, end, file_size, load_pickled, save_gt, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading graph from",file_name_in)
    g = load_pickled(file_name_in)
    end('')
    file_size(file_name_in)
    start("Initialzing id mapping and neighbours map")
    h = transform_nx_gt(g,rescale)
    end()
    start("Saving to",file_name_out)
    save_gt(file_name_out,h)
    end('')
    file_size(file_name_out)

def transform_nx_npz(g,rescale=False):
    from numpy import array, int64, int32, int16, float64, float32
    from limic.util import haversine_distance
    vs = g.nodes()
    es = g.edges.data(data=True)
    ids = [None]*len(vs)
    lat = [None]*len(vs)
    long = [None]*len(vs)
    id2edges = []
    edges_air = []
    edges_weight = []
    edges_neighbor = []
    v2index = {}
    index = 0
    ns = []
    for v in vs:
        v2index[v] = index
        ids[index] = v[0]
        lat[index] = int(v[1]*10000000) if rescale else v[1]
        long[index] = int(v[2]*10000000) if rescale else v[2]
        index += 1
        ns.append([])
    for e in es:
        w = e[2]['weight']
        air = e[2]['type']
        if rescale and air < 0:
            w = int(1000*haversine_distance(longx=e[0][2],latx=e[0][1],longy=e[1][2],laty=e[1][1]))
        f = v2index[e[0]]
        t = v2index[e[1]]
        ns[f].append((t,w,air))
        ns[t].append((f,w,air))
    counter = 0
    for n in ns:
        id2edges.append(counter)
        edges_weight.extend(map(lambda x:x[1],n))
        edges_neighbor.extend(map(lambda x:x[0],n))
        edges_air.extend(map(lambda x:x[2],n))
        counter += len(n)
    id2edges.append(counter)
    h = {}
    h['ids'] = array(ids,dtype=int64)
    if rescale:
        h['lat'] = array(lat,dtype=int32)
        h['long'] = array(long,dtype=int32)
        h['edges_weight'] = array(edges_weight,dtype=int32)
    else:
        h['lat'] = array(lat,dtype=float64)
        h['long'] = array(long,dtype=float64)
        h['edges_weight'] = array(edges_weight,dtype=float32)
    h['id2edges'] = array(id2edges,dtype=int32)
    h['edges_type'] = array(edges_air,dtype=int16)
    h['edges_neighbor'] = array(edges_neighbor,dtype=int32)
    return h

def convert_nx_npz(file_name_in,file_name_out,indirect=False,rescale=False,penalize=20):
    from limic.util import start, end, file_size, load_pickled, save_npz, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading graph from",file_name_in)
    g = load_pickled(file_name_in)
    end('')
    file_size(file_name_in)
    if indirect:
        start("Transforming to GT format")
        i = transform_nx_gt(g,rescale)
        end()
        start("Transforming to NPZ format")
        h = transform_gt_npz(i,penalize)
        end()
    else:
        start("Transforming to NPZ format")
        h = transform_nx_npz(g,rescale)
        end()
    start("Saving to",file_name_out)
    save_npz(file_name_out,h)
    end('')
    file_size(file_name_out)            

def convert_gt_nx(file_name_in,file_name_out,penalize=20):
    from limic.util import start, end, file_size, load_gt, save_pickled, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading graph from",file_name_in)
    g = load_gt(file_name_in)
    end('')
    file_size(file_name_in)
    start("Transforming to NPZ format")
    i = transform_gt_npz(g,penalize)
    end()
    start("Transforming to NX format")
    h = transform_npz_nx(i,penalize)
    end()
    start("Saving to",file_name_out)
    save_pickled(file_name_out,h)
    end('')
    file_size(file_name_out)            

def transform_gt_npz(g,penalize=20):
    from numpy import array, int64, int32, int16, float64, float32
    vs = g.get_vertices([g.vp.id,g.vp.lat,g.vp.long])
    rescale = g.gp.rescaled
    es = g.get_edges([g.ep.weight,g.ep.type])
    ids = [None]*len(vs)
    lat = [None]*len(vs)
    long = [None]*len(vs)
    id2edges = []
    edges_weight = []
    edges_neighbor = []
    edges_air = []
    ns = []
    for v in vs:
        index = v[0] if rescale else int(v[0])
        ids[index] = v[1]
        lat[index] = v[2]/(10000000. if rescale else 1)
        long[index] = v[3]/(10000000. if rescale else 1)
        ns.append([])
    for e in es:
        w = e[2]/((1000.*(penalize if e[3] else 1)) if rescale else 1)
        air = e[3]
        f = e[0] if rescale else int(e[0])
        t = e[1] if rescale else int(e[1])
        ns[f].append((t,w,air))
        ns[t].append((f,w,air))
    counter = 0
    for n in ns:
        id2edges.append(counter)
        edges_weight.extend(map(lambda x:x[1],n))
        edges_neighbor.extend(map(lambda x:x[0],n))
        edges_air.extend(map(lambda x:x[2],n))
        counter += len(n)
    id2edges.append(counter)
    h = {}
    h['ids'] = array(ids,dtype=int64)
    h['lat'] = array(lat,dtype=float64)
    h['long'] = array(long,dtype=float64)
    h['id2edges'] = array(id2edges,dtype=int32)
    h['edges_weight'] = array(edges_weight,dtype=float32)
    h['edges_neighbor'] = array(edges_neighbor,dtype=int32)
    h['edges_type'] = array(edges_air,dtype=int16)
    return h

def convert_gt_npz(file_name_in,file_name_out,penalize=20):
    from limic.util import start, end, file_size, load_gt, save_npz, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading graph from",file_name_in)
    g = load_gt(file_name_in)
    end('')
    file_size(file_name_in)
    start("Initialzing id mapping and neighbours map")
    h = transform_gt_npz(g,penalize)
    end()
    start("Saving to",file_name_out)
    save_npz(file_name_out,h)
    end('')
    file_size(file_name_out)

def transform_npz_nx(g,penalize=20):
    from numpy import int32
    ids = g['ids']
    lat = g['lat']
    long = g['long']
    id2edges = g['id2edges']
    edges_weight = g['edges_weight']
    edges_neighbor = g['edges_neighbor']
    edges_air = g['edges_type']
    rescale = edges_weight.dtype == int32
    from networkx import Graph
    h = Graph()
    index2node = {}
    for index in range(len(ids)):
        node = (int(ids[index]),lat[index]/(10000000. if rescale else 1),long[index]/(10000000. if rescale else 1))
        index2node[index] = node
        h.add_node(node)
    for index in range(len(ids)):
        left = id2edges[index]
        right = id2edges[index+1]
        me = index2node[index]
        for i in range(left,right):
            neighbor = edges_neighbor[i]
            air = edges_air[i]
            w = edges_weight[i]/((1000.*(penalize if air < 0 else 1)) if rescale else 1)
            h.add_edge(me,index2node[neighbor],weight=w,type=air)
    return h

def convert_npz_nx(file_name_in,file_name_out,penalize=20):
    from limic.util import start, end, file_size, save_pickled, load_npz, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading graph from",file_name_in)
    g = load_npz(file_name_in)
    end('')
    file_size(file_name_in)
    start("Transforming to NX format")
    h = transform_npz_nx(g,penalize)
    end()
    start("Saving to",file_name_out)
    save_pickled(file_name_out,h)
    end('')
    file_size(file_name_out)            

def convert_npz_gt(file_name_in,file_name_out,rescale=False,penalize=20):
    from limic.util import start, end, file_size, save_gt, load_npz, check_overwrite
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading graph from",file_name_in)
    g = load_npz(file_name_in)
    end('')
    file_size(file_name_in)
    start("Transforming to NX format")
    i = transform_npz_nx(g,penalize)
    end()
    start("Transforming to GT format")
    h = transform_nx_gt(i,rescale)
    end()
    start("Saving to",file_name_out)
    save_gt(file_name_out,h)
    end('')
    file_size(file_name_out)            

def convert_cache_dbm(file_name_in,file_name_out):
    from limic.util import start, end, file_size, status, load_pickled, check_overwrite
    from dbm.gnu import open as dopen
    from os.path import exists
    from pickle import dumps
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Loading from",file_name_in)
    d = load_pickled(file_name_in)
    end('')
    file_size(file_name_in)
    start("Opening database",file_name_out)
    if not exists(file_name_out):
        db = dopen(file_name_out,"c")
        db.close()
    db = dopen(file_name_out,"c")
    end('')
    file_size(file_name_out)
    start("Computing set of entries to save")
    for key in db.keys():
        del d[key.decode("utf-8")]
    status(len(d))
    start("Saving entries to",file_name_out)
    for key, val in d.items():
        db[key.encode("utf-8")] = dumps(val)
    db.close()
    end('')
    file_size(file_name_out)

def convert_dbm_cache(file_name_in,file_name_out):
    from limic.util import start, end, file_size, status, save_pickled, check_overwrite
    from dbm.gnu import open as dopen
    from pickle import loads
    if not check_overwrite(file_name_in,file_name_out):
        return
    start("Opening database",file_name_in)
    db = dopen(file_name_in,"r")
    end('')
    file_size(file_name_in)
    start("Converting to dictionary")
    d = {}
    for key in db.keys():
        d[key.decode("utf-8")] = loads(db[key])
    db.close()
    end()
    start("Saving to",file_name_out)
    save_pickled(file_name_out,d)
    end('')
    file_size(file_name_out)



from distutils.core import setup

setup(
    name='LiMiC',
    version='0.3.5',
    packages=['limic',],
    package_data={'limic':['render.js']},
    include_package_data=True,
    license='MIT License',
    url='https://pypi.org/project/LiMiC/',
    author='Peter Schneider-Kamp',
    author_email='petersk@imada.sdu.dk',
    description='Linear-infrastructure Mission Control (LiMiC)',
    long_description=open('README.txt').read(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['requests','numpy','dogpile.cache','networkx','shapely','folium','flask','scipy','pyproj'],
    python_requires='>=3.4'
)



GRAPH = (("file_name",),{'type':str,'help':"use graph file GRAPH",'metavar':'GRAPH'})
CONFIG = [
    ("nx",{'help':"count total length of edges in NX graph",'args':[GRAPH]}),
    ("gt",{'help':"count total length of edges in GT graph",'args':[GRAPH]}),
    ("npz",{'help':"count total length of edges in NPZ graph",'args':[GRAPH]})
]

def compute_length_gt(g):
    es = g.get_edges([g.ep.weight,g.ep.type])
    total_length = 0
    for e in es:
        if e[3] >= 0:
            total_length += e[2]/1000.
    return total_length

def length_nx(file_name):
    from limic.util import start, end, status, load_pickled
    from limic.convert import transform_nx_gt
    start("Loading graph from",file_name)
    g = load_pickled(file_name)
    end()
    start("Transforming graph to rescaled GT format")
    h = transform_nx_gt(g,rescale=True)
    end()
    start("Computing length using rescaled GT")
    length = compute_length_gt(h)
    end('')
    status(length)

def length_gt(file_name):
    from limic.util import start, end, status, load_gt
    from limic.convert import transform_gt_npz, transform_npz_nx, transform_nx_gt
    start("Loading graph from",file_name)
    g = load_gt(file_name)
    end()
    start("Checking whether GT graph is rescaled")
    if g.gp.rescaled:
        status("YES")
        start("Computing length using GT")
        length = compute_length_gt(g)
        end('')
        status(length)
    else:
        status("NO (forcing reconversion)")
        start("Transforming graph to NPZ format")
        h = transform_gt_npz(g,penalize=20)
        end()
        start("Transforming graph to NX format")
        i = transform_npz_nx(h)
        end()
        start("Transforming graph to rescaled GT format")
        j = transform_nx_gt(i,rescale=True)
        end()
        start("Computing length using rescaled GT")
        length = compute_length_gt(j)
        end('')
        status(length)

def length_npz(file_name):
    from limic.util import start, end, status, load_npz
    from limic.convert import transform_npz_nx, transform_nx_gt
    start("Loading graph from",file_name)
    g = load_npz(file_name)
    end()
    start("Transforming graph to NX format")
    h = transform_npz_nx(g)
    end()
    start("Transforming graph to rescaled GT format")
    i = transform_nx_gt(h,rescale=True)
    end()
    start("Computing length using rescaled GT")
    length = compute_length_gt(i)
    end('')
    status(length)



COUNTRIES = (("countries",),{'type':str,'nargs':'+','help':"countries to route on",'metavar':'COUNTRY'})
GRAPH = (("graph_file",),{'type':str,'help':"use graph file GRAPH",'metavar':'GRAPH'})
HTML = (("html_file",),{'type':str,'help':"use rendered HTML file",'metavar':'HTML'})
RHOST = (("--render-host",),{'type':str,'dest':'r_host','default':'localhost','help':"render hostname (default: localhost)",'metavar':'HOST'})
RPORT = (("--render-port",),{'type':int,'dest':'r_port','default':5000, 'help':"render port number (default: 5000)",'metavar':'PORT'})
RPREFIX = (("--render-prefix",),{'type':str,'dest':'r_prefix','default':'','help':"render URI prefix (default: '')",'metavar':'PREFIX'})
HOST = (("--host",),{'type':str,'dest':'host','default':'localhost','help':"hostname (default: localhost)",'metavar':'HOST'})
PORT = (("--port",),{'type':int,'dest':'port','default':5000, 'help':"port number (default: 5000)",'metavar':'PORT'})
PREFIX = (("--prefix",),{'type':str,'dest':'prefix','default':'','help':"URI prefix (default: '')",'metavar':'PREFIX'})
OSM = (("--osm",),{'action':'store_true','dest':'osm','default':False,'help':"process from OSM file (SLOW)"})
NOBROWSER = (("-n","--no-browser"),{'action':'store_true','dest':'no_browser','default':False,'help':"do not open html in browser (headless)"})
URL = (("-d","--download-url"),{'type':str,'dest':'url','help':"define url for download directory to be URL",'metavar':'URL'})
SHOW = (("-l","--list"),{'action':'store_true','dest':'show','default':False,'help':"list available countries (default: False)"})
CONSERVEMEM = (("-c","--conserve-memory"),{'action':'store_true','dest':'conserve_mem','default':False,'help':"lower memory usage but higher runtime"})

CONFIG = [
    ("auto",{'help':"start HTTP server for given COUNTRY (default: Europe)",'args':[SHOW,URL,CONSERVEMEM,NOBROWSER,HOST,PORT,PREFIX,RHOST,RPORT,RPREFIX,OSM,COUNTRIES]}),
    ("nx",{'help':"start HTTP server for given GRAPH and rendered HTML file",'args':[SHOW,URL,NOBROWSER,HOST,PORT,PREFIX,RHOST,RPORT,RPREFIX,GRAPH,HTML]}),
]

def serve_auto(countries,host="localhost",port=5000,prefix="",osm=False,url=None,show=False,no_browser=False,r_host="localhost",r_port=5000,r_prefix="",conserve_mem=False):
    from limic.download import common, download_graph, download_merged, download_osm
    from limic.init import extract_osm_all, merge_all
    from limic.render import render_nx
    graph_file = None
    countries, url = common(countries=countries,url=url,show=show,osm=osm)
    if osm:
        download_osm(countries=countries,url=url)
        extract_osm_all(countries=countries,conserve_mem=conserve_mem)
    else:
        download_graph(suffix="nx",countries=countries,url=url,show=show)
    if len(countries) > 1 :
        merge_all(countries)
    if len(countries) == 1:
        graph_file = "graph."+countries[0]+".nx"
    html_file = graph_file[:-2]+"html"
    serve_nx(graph_file,html_file,host=host,port=port,prefix=prefix,url=url,show=show,no_browser=no_browser,r_host=r_host,r_port=r_port,r_prefix=r_prefix)
        
    
def serve_nx(graph_file,html_file,host="localhost",port=5000,prefix="",url=None,show=False,no_browser=False,r_host="localhost",r_port=5000,r_prefix=""):
    from flask import Flask, jsonify, request
    from limic.util import load_pickled, start, end, status
    from limic.route import astar_nx
    from limic.render import render_nx
    from threading import Thread
    from time import sleep
    from webbrowser import open as wopen
    from scipy.spatial import cKDTree as KDTree
    from pyproj import CRS, Transformer
    g = render_nx(graph_file,html_file,host=r_host,port=r_port,prefix=r_prefix)
    start("Initializing KD-Tree")
    nodes = list(g.nodes())
    crs_4326 = CRS("WGS84")
    crs_proj = CRS("EPSG:28992")
    transformer = Transformer.from_crs(crs_4326, crs_proj)
    tree = KDTree(list(map(lambda x:transformer.transform(x[1],x[2]),nodes)))
    end()

    start("Setting up the app")
    app = Flask("LiMiC")

    @app.route(prefix+"/")
    def hello():
        return open(html_file).read()

    @app.route(prefix+"/tower")
    def tower():
        lat = float(request.args.get('lat'))
        lng = float(request.args.get('lng'))
        start("Finding tower",lat,lng)
        tower = nodes[tree.query(transformer.transform(lat,lng))[1]]
        end('')
        res = jsonify(tower=tower)
        end()
        return res

    @app.route(prefix+"/route")
    def route():
        source_lat = float(request.args.get('source[lat]'))
        source_lng = float(request.args.get('source[lng]'))
        target_lat = float(request.args.get('target[lat]'))
        target_lng = float(request.args.get('target[lng]'))
        start("Routing",source_lat,source_lng,target_lat,target_lng)
        source = nodes[tree.query(transformer.transform(source_lat,source_lng))[1]]
        end('')
        target = nodes[tree.query(transformer.transform(target_lat,target_lng))[1]]
        end('')
        path = astar_nx(g,source,target)
        end('')
        if path[1][-1][0] == float('inf'):
            path[1][-1] = (path[1][-1][1],)+path[1][-1][1:]
        res = jsonify(path=path)
        end()
        return res
    end()

    class OpenThread(Thread):
        def run(self):
            delay = 0.5
            sleep(delay)
            url = "http://%s:%d%s/" % (host,port,prefix)
            start("Opening",url,"in browser")
            wopen(url,new=2)
            status("DONE")
    if not no_browser:
        OpenThread().start()
    app.run(host=host,port=port)
